function [ data, g, data0 ] = ...
    projReachOneGen(t_max, g, data0, p, restrict_update, accuracy)
% projReachOneGen: DAE reach set for a simple power system model.
%
%   [ data, g, data0 ] = ...
%               projReachOneGen(t_max, g, data0, p, restrict_update, accuracy)
%  
% This function computes reachability for a "rudimentary but
% representative model of the power system" described in:
%
%   Venkatasubramanian, Schattler & Jaborszky, "Voltage Dynamics: Study
%   of a Generator with Voltage Control, Transmission, and Matched MW
%   Load," IEEE Trans. Automatic Control, v. 37, n. 11, pp. 1717-1733
%   (November, 1992).
%
% The model consists of a single generator and single load, connected by a
% transmission line.  This DAE model was brought to my attention by
% Yoshihiko Susuki at Kyoto University, Dept. Electrical Engineering, who
% performed preliminary reachability analysis using trajectories; see
%
%   Susuki & Hikihara, "Predicting Voltage Instability of Power System via
%   Hybrid System Reachability Analysis," Proc. ACC, pp. 4166-4171, New
%   York, NY (2007).
%
% This 3D model has state vector (E', E_fd, E), where evolution of the first
% two are governed by DEs and the final variable E solves an algebraic
% constraint.  The model is given by (10)--(12) in Venkatasubramanian.
%
% The reach set is tracked in a two dimensional space corresponding to the
% manifold on which the DAE evolves.  Because the grid is two dimensional,
% we cannot create implicit surface functions for the constraint surface
% itself.
%
% This routine differs from the one of the same name used in the ACC08
% submission because its input parameters include a final time, a grid, an
% initial condition implicit surface, and a model parameter structure
% (instead of requiring the model parameters to be defined locally).  Unlike
% that routine, it returns only the terminal implicit surface function.
%
% This function was originally designed as a script file, so most of the
% options can only be modified in the file.  For example, edit the file to
% change the grid dimension, boundary conditions, flow field parameters,
% etc.
%
% Input Parameters:
%
%   t_max: Positive double.  Time horizon for the reachability calculation.
%   Optional.  Defaults to 5 time units.
%
%   g: Grid structure.  Specifies the computational grid on which to run
%   the reachability.  First variable is E, second is E_fd.  Optional.
%   You may use the empty matrix [] as a placeholder if you wish to use
%   the default grid.
%
%   data0: Array of size appropriate to the grid.  Contains an implicit
%   surface function which is used as the target set for the reachability
%   calculation.  Optional.  Defaults to the unsafe set (E <= E_min).
%   You may use the empty matrix [] as a placeholder if you wish to use
%   the default initial data set.
%
%   p: Structure.  Parameter structure for the electrical system model.
%   Note that if you use something other than the default parameters, the
%   grid may have to change (eg bounds may be wrong).  Optional.  Defaults
%   to the parameter values from figures 4 & 8 of Venkatasubramanian.
%   You may use the empty matrix [] as a placeholder if you wish to use
%   the default parameters.
%
%   restrict_update: Boolean.  Specifies whether the reach set should only
%   grow.  Optional.  Default is 1.
%
%   accuracy: String.  Controls the order of approximations.
%
%     'low': Use odeCFL1 and upwindFirstFirst.
%     'medium': Use odeCFL2 and upwindFirstENO2 (default).
%     'high': Use odeCFL3 and upwindFirstENO3.
%     'veryHigh': Use odeCFL3 and upwindFirstWENO5.
%
% Output Parameters:
%
%   data: Array containing the implicit surface function at t_max.
%
%   g: Grid structure on which data was computed.
%
%   data0: Implicit surface function at t_0.

% Copyright 2007 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Ian Mitchell, 10/3/07

  %---------------------------------------------------------------------------
  if(nargin < 1)
    t_max = 5;
  end

  if(nargin < 2)
    % Signal that we need a grid structure.
    g = [];
  end
  
  if(nargin < 3)
    data0 = [];
  end
  
  if(nargin < 4)
    p = [];
  end
  
  if(nargin < 5)
    restrict_update = 1;
  end

  if(nargin < 6)
    accuracy = 'medium';
  end
    
  %---------------------------------------------------------------------------
  % You will see many executable lines that are commented out.
  %   These are included to show some of the options available; modify
  %   the commenting to modify the behavior.

  % What kind of display?  Used only in the calls to visualizeLevelSet; the
  % open curves are visualized separately.
  displayType = 'contour';

  % The E_fd dimension is much larger than the other two, so we need to scale
  % the visualizations.  Instead, however, we will scale the coordinates,
  % which implicitly scales both the visualization and the computation.
  aspectRatio = [ 1, 1, 1 ];
  scaleFactor = [ 1; 1; 0.1 ];

  %---------------------------------------------------------------------------
  % Model parameters.
  if isempty(p)
  
    % How many transmission routes?  Parameters are designed for 1 or 2.
    % Venkatasubramanian uses 2 throughout.
    p.valid_routes = [ 1; 2 ];
    p.num_routes = 2;

    % Problem parameters from Venkatasubramanian.
    p.Tp_d0 = 5;
    p.E_r   = 1.0;
    p.H     = 0;
    p.x_d   = 1.2;
    p.B     = 0;
    p.xp_d  = 0.2;
    p.K     = 7.0;

    % Parameters from figure 4.
    p.T     = 1.5;
    p.E0_fd = 1.6;
    % Choice of P determines which subfigure of figure 8 is in effect.
    p.P     = 1.0;
  
    % Calculated parameters.
    p.x     = [ 0.2; 0.1 ];
    p.xp    = p.x + p.xp_d;
  
    p.Q0    = 0.5 * p.P;
    if((p.H == 0) && (p.B == 0))
      % By default, we will have the same Q in both old and new 
      % configurations: no extra Q factor added in.
      p.Q = p.Q0;
    else
      error('DAE model is not currently set up to handle nonconstant Q');

      % The problem is that Q will now depend on state variable E, so the 
      % code in the dynamics routine would have to be modified.  Not a 
      % huge deal for the DAE model, but the fully ODE model would have 
      % to be rederived.
      %
      % Q = p.Q0 + p.H * E + p.B * E.^2
    end

    % Reachability parameters.  Also appears as E_c in the papers.
    p.E_min = 0.70;
  end
  
  %---------------------------------------------------------------------------
  % Integration parameters.
  plotSteps = 11;              % How many intermediate plots to produce?
  t0 = 0;                      % Start time.
  singleStep = 0;              % Plot at each timestep (overrides t_plot).

  % Period at which intermediate plots should be produced.
  t_plot = (t_max - t0) / (plotSteps - 1);
  
  % How close (relative) do we need to get to t_max to be considered finished?
  small = 100 * eps;

  %---------------------------------------------------------------------------
  % What level set should we view?
  level = 0;

  % How many steps of reinitialization after each timestep?  Set to zero for
  % no reinitialization.  Empirically, one reinit step seems to help (I have
  % not tried two).
  reinitSteps = 0;

  % Pause after each plot?
  pauseAfterPlot = 0;

  % Delete previous plot before showing next?
  deleteLastPlot = 0;

  % Plot in separate subplots (set deleteLastPlot = 0 in this case)?
  useSubplots = 0;

  %---------------------------------------------------------------------------
  % Use these indices on the problem coordinate cell vector and other
  % things that work in the full dimensional formulation.
  E_index = 1;
  Ep_index = 2;
  E_fd_index = 3;

  if isempty(g)
    % Create the computational grid.  Our coordinates correspond to E and
    % E_fd in the original full dimensional problem.
    g.dim = 2;
    % Reasonably fast runs use 2, figure in paper used 4.
    g.N = 2 * [ 60; 40 ] + 1;
    g.min = [ 0.5; -1 ] .* scaleFactor([ E_index, E_fd_index ]);
    g.max = [ 1.7; +7 ] .* scaleFactor([ E_index, E_fd_index ]);
    g.bdry = @addGhostExtrapolate;
    g = processGrid(g);
  end

  % Computational dimensions z_1 and z_2 are mapped to the original problem
  % coordinates by the subfunction coordinateMap().
  problem.dim = 3;
  problem.xs = coordinateMap(g.xs, p);

  E = problem.xs{E_index} / scaleFactor(E_index);
  Ep = problem.xs{Ep_index} / scaleFactor(Ep_index);
  E_fd = problem.xs{E_fd_index} / scaleFactor(E_fd_index);

  %--------------------------------------------------------------------------- 
  % Choose approximations at appropriate level of accuracy.
  switch(accuracy)
   case 'low'
    derivFunc = @upwindFirstFirst;
    integratorFunc = @odeCFL1;
   case 'medium'
    derivFunc = @upwindFirstENO2;
    integratorFunc = @odeCFL2;
   case 'high'
    derivFunc = @upwindFirstENO3;
    integratorFunc = @odeCFL3;
   case 'veryHigh'
    derivFunc = @upwindFirstWENO5;
    integratorFunc = @odeCFL3;
   otherwise
    error('Unknown accuracy level %s', accuracy);
  end

  %---------------------------------------------------------------------------
  if isempty(data0)
    % Create initial conditions: Plane at E = E_min represents too low a bus
    % voltage.  Negative values inside unsafe set.  We need to multiply the
    % result by the scaleFactor to get a signed distance function.
    data0 = (E - p.E_min) * scaleFactor(E_index);
  end

  data = data0;

  %---------------------------------------------------------------------------
  % Choose the flow field.

  % Basic DAE differential equations.
  E_G = sqrt((p.x(p.num_routes) * p.P)^2 ...
             + (p.x(p.num_routes) * p.Q + E.^2).^2) ./ E;
  Ep_dot = (E_fd -(p.x(p.num_routes) + p.x_d) / p.xp(p.num_routes) * Ep ...
            + ((p.x_d - p.xp_d) / p.xp(p.num_routes)) ...
            * (E.^2 + p.xp(p.num_routes) * p.Q) ./ Ep) / p.Tp_d0;
  
  E_fd_dot = -(1/p.T) * ((E_fd - p.E0_fd) + p.K * (E_G - p.E_r));

  % Used to map onto the manifold.  This equation relates Ep to the manifold
  % coordinate system.  Since E_fd is one of the manifold coordinates,
  % there is no need to map its dynamics.
  sigmaInv = (E.^2 .* sqrt((p.xp(p.num_routes) * p.P)^2 ...
                           + (p.xp(p.num_routes) * p.Q + E.^2).^2) ...
              ./ (E.^4 - (p.xp(p.num_routes) * p.P)^2 ...
                  - (p.xp(p.num_routes) * p.Q)^2));

  % We will damp out the dynamics within the target set to avoid singularity
  % in the coordinate mapping.  There is no need to get signed distance for
  % the target set implicit surface function because it is already signed
  % distance.  Damping function is one for all positive values of the target
  % set's implicit surface function, and then damps smoothly down to
  % zero.

  % Range over which the damping is gradually applied should depend on
  % the grid resolution; in particular, since the target set depends
  % entirely on E, we need to pick out that grid resolution.
  damp_distance = 2 * g.dx(E_index);
  % Useful intermediate computation.
  scale_shift_data0 = (data0 + damp_distance) / damp_distance;

  damping = ((data0 > 0) ...
             + (((data0 <= 0) & (data0 > -2 * damp_distance)) ...
                .* (0.5 + (0.5 * scale_shift_data0) ...
                    + 0.5 / pi * sin(pi * scale_shift_data0))));

  % Velocity field.  Multiply by negative one to reverse time.  Multiply by
  % damping to avoid the singular surface inside the target set.  Also,
  % scale the dynamics in line with the coordinate scaling.
  v = { -scaleFactor(E_index) * damping .* sigmaInv .* Ep_dot; ...
        -scaleFactor(E_fd_index) * damping .* E_fd_dot };
  
  % Set up the convection routine.
  schemeFunc = @termConvection;

  schemeData.velocity = v;
  schemeData.derivFunc = derivFunc;
  schemeData.grid = g;

  %---------------------------------------------------------------------------
  % Set up time approximation scheme.
  integratorOptions = odeCFLset('factorCFL', 0.9, 'stats', 'on');

  if(singleStep)
    integratorOptions = odeCFLset(integratorOptions, 'singleStep', 'on');
  end
  
  %---------------------------------------------------------------------------
  % Restrict the Hamiltonian so that reachable set only grows.
  %   The regular approximation scheme MUST already be completely set up.
  if(restrict_update)
    innerFunc = schemeFunc;
    innerData = schemeData;
    clear schemeFunc schemeData;
    
    % Wrap the dynamics inside the term approximation restriction routine.
    schemeFunc = @termRestrictUpdate;
    schemeData.innerFunc = innerFunc;
    schemeData.innerData = innerData;
    schemeData.positive = 0;
  end
  
  %---------------------------------------------------------------------------
  % Set up reinitialization if necessary.
  if(reinitSteps)
    % Register reinitialization postTimestep routine.
    integratorOptions = odeCFLset(integratorOptions, ...
                                  'PostTimestep', @postTimestepReinit);
    
    % Add necessary components to schemeData structures.
    schemeData.reinitAccuracy = accuracy;
    schemeData.reinitSteps = reinitSteps;
    % We also need the grid structure, in case we are using restrict_update.
    schemeData.grid = g;
  end

  %---------------------------------------------------------------------------
  % Initialize Display
  f = figure;

  % Set up subplot parameters if necessary.
  if(useSubplots)
    rows = ceil(sqrt(plotSteps));
    cols = ceil(plotSteps / rows);
    subplot(rows, cols, plotNum);
  end
  
  % Show the evolving level set.
  h = visualizeLevelSet(g, data, displayType, level, [ 't = ' num2str(t0) ]);
  hold on;
  
  %---------------------------------------------------------------------------
  % Loop until t_max (subject to a little roundoff).
  plot_num = 1;
  t_now = t0;
  startTime = cputime;
  while(t_max - t_now > small * t_max)

    % Reshape data array into column vector for ode solver call.
    y0 = data(:);

    % How far to step?
    tSpan = [ t_now, min(t_max, t_now + t_plot) ];
  
    % Take a timestep.
    [ t, y, schemeData ] = feval(integratorFunc, schemeFunc, tSpan, y0,...
                                 integratorOptions, schemeData);
    t_now = t(end);

    % Get back the correctly shaped data array
    data = reshape(y, g.shape);

    if(pauseAfterPlot)
      % Wait for last plot to be digested.
      pause;
    end

    % Delete last visualization if necessary.
    if(deleteLastPlot)
      delete(h);
    end

    % Move to next subplot if necessary.
    if(useSubplots)
      subplot(rows, cols, plot_num);
      hold on;
    end

    % Create new visualization.
    h = visualizeLevelSet(g, data, displayType, level,['t = ' num2str(t_now)]);

    axis(g.axis);
    daspect(aspectRatio);
    drawnow; 
  
    plot_num = plot_num + 1;
    
  end

  endTime = cputime;
  fprintf('Total execution time %g seconds\n', endTime - startTime);



%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function xsOut = coordinateMap(xsIn, params)
% coordinateMap: maps computational coordinates to problem coordinates.
%
%   xsOut = coordinateMap(xsIn, params)
%
% Converts between 2D computational coordinates and 3D problem coordinates
% for the one generator example.  If x = ( x_1, x_2 ) are the computational
% coordinates, then the problem coordinates E_fd, E, E' are found by:
%
%    E    = x_1
%    E'   = sqrt((x' P)^2 + (x' Q + E^2)^2) / E
%    E_fd = x_2
%
% Input Parameters:
%
%   xsIn: the 2D grid.xs cell vector giving the location of the
%   computational nodes.
%
%   params: A structure containing the problem parameters.
%
% Output Parameters:
%
%   xsOut: A 3D cell vector giving the location in the problem coordinate
%   system of the computational nodes.
  
  xsOut = cell(3,1);

  % E.
  E = xsIn{1};
  xsOut{1} = E;

  % E'.
  xsOut{2} = sqrt((params.xp(params.num_routes) * params.P)^2 + ...
                  (params.xp(params.num_routes) * params.Q + E.^2).^2) ./ E;
  
  % E_fd.
  xsOut{3} = xsIn{2};
