function new_data = resetOneGen(old_data, new_routes, g, p, accuracy)
% resetOneGen: Applies the reset map when transmission routes change.
%
%   new_data = resetOneGen(old_data, new_routes, g, p, accuracy)
%
% This routine applies the (reverse) reset mapping when the number of
% transmission routes change in the simple 3D DAE model of the power system
% from:
%
%   Venkatasubramanian, Schattler & Jaborszky, "Voltage Dynamics: Study
%   of a Generator with Voltage Control, Transmission, and Matched MW
%   Load," IEEE Trans. Automatic Control, v. 37, n. 11, pp. 1717-1733
%   (November, 1992).
%
% This routine is specialized to the reachability computation on the
% constraint manifold -- the 2D version of the reachability problem.
% Therefore, it operates in the E vs E_fd plane.  Note the the E' dimension
% is relevant only to the computation of the mapping (which is done by
% switchOneGen), not to the application of the mapping.
%
% The real trick to this routine is how to handle the grid nodes that do not
% have good pre-switch states: either the state before the switch is
% undefined (eg. the constraint cannot be satisfied for the given E'), the
% state before the switch is unsafe (inside the target set), or the state
% before the switch is outside the computational grid.  To deal with these
% cases, we build an approximate implicit surface function using some hacks,
% and then apply reinitialization to improve the quality of the implicit
% surface function.
%
% Input Parameters:
%
%   old_data: Array.  Contains the implicit surface function for the mode
%   before the switch.
%
%   new_routes: Integer.  The number of transmission routes in the system
%   after the switch.  Passed to switchOneGen.
%
%   g: Computational grid structure.  Used for both the pre and post
%   switch functions.
%
%   p: Structure.  Parameter structure for the electrical system model.
%   Note that if you use something other than the default parameters, the
%   visualization may not look very good (eg bounds may be wrong).
%
%   accuracy: String.  Controls the order of approximations in the
%   reinitialization.  Note that linear interpolation is always used for the
%   actual reset mapping.
%
%     'none': Do not reinitialize.
%     'low': Use odeCFL1 and upwindFirstFirst.
%     'medium': Use odeCFL2 and upwindFirstENO2 (default).
%     'high': Use odeCFL3 and upwindFirstENO3.
%     'veryHigh': Use odeCFL3 and upwindFirstWENO5.
%
% Output Parameters:
%
%   new_data: Array. Contains the implicit surface function for the mode
%   after the switch.

% Copyright 2007 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Ian Mitchell, 10/08/07

  
  if(nargin < 5)
    accuracy = 'medium';
  end

  %-------------------------------------------------------------------------
  % We need to sample the implicit surface function on the new mode's grid.
  % Watch out for scaling factors that may have been introduced.
  scaling_E = 1;

  % Extract the grid nodes for the E dimension.
  samples_E = g.vs{1} / scaling_E;

  % Compute the effect of the reset.
  old_coords = switchOneGen(new_routes, 0, samples_E, p);

  %-------------------------------------------------------------------------
  % The grid nodes that have bad pre-switch states occur at the top and
  % bottom of the computational domain.
    
  % At the top of the domain, states may map outside the domain.  We know for
  % this problem that states outside the computational domain are unsafe in
  % every mode.  Therefore, if the reset maps outside the computational
  % domain, we will simply set the implicit surface function to the most
  % negative value we have encountered.
  before_switch_min = min(old_data(:));

  % The situation at the bottom of the domain is trickier.  To handle it, we
  % will fill in the implicit surface function value for nodes with valid
  % pre-switch coordinates first, and then come back to fill in the invalid
  % ones at the bottom of the domain.

  % Measure how many nodes in E dimension have invalid pre-switch states
  % at the bottom end of the domain.
  num_invalid = 0;

  % Flag to indicate that we are done with the "bottom end" of the domain.
  found_valid = 0;

  %-------------------------------------------------------------------------
  % Coordinates for the grid nodes in the E_fd dimension.
  interp_E_fd = g.vs{2}';
  
  % Storage space for the new data.  We will fill it with NaN to check
  % that all nodes are given a valid value by the end of the routine.
  new_data = nan * ones(size(old_data));
  
  % Work one E value at a time through the grid.
  for i = 1 : length(samples_E)
    
    if isnan(old_coords(i, 1))
      % Invalid pre-switch coordinate.

      if found_valid
        % We are at the top of the domain.  Use most negative value 
        % encountered.
        new_data(i, :) = before_switch_min;
        
      else
        % We are at the bottom of the domain.  We will deal with these
        % nodes later.
        num_invalid = num_invalid + 1;
      end
      
    else
      % The pre-switch coordinate is valid.
      found_valid = 1;

      % Find the implicit surface function value at the pre-switch
      % coordinate location.
      interp_E = old_coords(i, 1) * ones(size(interp_E_fd));
      new_data(i,:) = interpn(g.xs{:}, old_data, interp_E, interp_E_fd);

    end
    
  end

  %-------------------------------------------------------------------------
  % Now we can go back and deal with those nodes without valid pre-switch
  % coordinates at the bottom of the domain.  We will have more nodes in
  % this situation and we don't want a broad swath of the implicit surface
  % function to be constant.  Therefore, we will try to slope away from the
  % implicit surface function via two measures.  First, we will measure how
  % far the node is from the left-most node with a valid pre-switch state.
  % Second, all nodes will be given a value less than that of neighbours
  % with valid pre-switch states.
  
  % We will work backwards from the first node with valid pre-switch state to
  % the bottom of the domain so that we can try to keep the implicit surface
  % function sloping down away from the implicit surface.
  for i = num_invalid : -1  : 1
    
    % The first measure: A value which gradually slopes away from the implicit
    % surface.
    unsafe_value = -g.dx(1) * (num_invalid - i + 1);
    
    % The second measure: More negative than all neighbors.  In
    % particular, the neighbors to the right, which should by now have a
    % valid implicit surface function value.
    less_than_neighbor = new_data(i + 1, :) - g.dx(1);
    
    % Take the minimum of the two measures.
    new_data(i,:) = min(unsafe_value, less_than_neighbor);
    
  end
  
  %-------------------------------------------------------------------------
  % We've done our best to construct a reasonable implicit surface function,
  % but it will still have some issues -- kinks, flat spots, etc.  Use a
  % little bit of reinitialization to make those nicer.  
  
  % We will run the reinit for a number of iterations twice the number of
  % invalid nodes we had at the bottom of the domain; that should be
  % sufficient to sweep out any messiness at that end of the domain, and
  % should be more than enough to handle the top of the domain.
  if ~strcmp(lower(accuracy), 'none')
    new_data = signedDistanceIterative(g, new_data, accuracy, ...
                                       -2 * num_invalid);
  end
