% runHybridOneGen: A script for the HSCC 2008 example.
%
% This script file recreates the reachable tube calculation from
%
%    Ian M. Mitchell & Yoshihiko Susuki, "Level Set Methods for Computing
%    Reachable Sets of Hybrid Systems with Differential Algebraic
%    Equation Dynamics," submitted to Hybrid Systems Computational &
%    Control 2008.
%
% The actual evolution of the reachable tube is performed by projReachOneGen
% (for the continuous evolution) or resetOneGen (for the discrete
% evolution).  This script implements the switching logic for the hybrid
% automata (HA) and generates the figure.
%
% The parameters are taken from either Figures 4 & 8 of that paper or from a
% supplementary sheet of Susuki's HSCC 2006 poster "Application of Hybrid
% System Theory to Power System Voltage Stability Analysis."  The
% Venkatasubramanian parameters are preferred (and were used in both the ACC
% 2008 and HSCC 2008 submissions); if you want to use the Susuki parameters
% you will need to change the computational domain.  The only change from
% the ACC 2008 submission is Q_0 = 0.25 P_m (instead of 0.5 P_m).  With the
% ACC 2008 value, the safe set for the hybrid system was empty;
% consequently, I reduced Q_0 at Yoshihiko's suggestion to increase the safe
% set's size.

% Copyright 2007 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Ian Mitchell, 10/12/07

  %---------------------------------------------------------------------------
  % Make sure we can see the ToolboxLS kernel m-files.
  run('../addPathToKernel');

  %---------------------------------------------------------------------------

  % From what file should we load the data for visualization?  If this
  % constant is the empty matrix, then we will recalculate the data from
  % scratch.
  data_filename = [];

  % Accuracy of level set calculation.
  accuracy = 'medium';
  %accuracy = 'high';
  
  %---------------------------------------------------------------------------
  % Model parameters.

  % Use which parameter set?  The Cross & Mitchell ACC submission and Cross's
  % thesis (using closest point methods in the full state space) did not use
  % Susuki's parameters.  Note that the domain boundaries and/or integration
  % time may need to be adjusted when a different parameter set is chosen.
  useSusuki = 0;

  p.valid_routes = [ 1; 2 ];
  
  % Problem parameters common to both Susuki and Venkatasubramanian.
  p.Tp_d0 = 5;
  p.E_r   = 1.0;
  p.H     = 0;
  p.x_d   = 1.2;
  p.B     = 0;
  p.xp_d  = 0.2;
  p.K     = 7.0;

  % Problem parameters that differ.
  if(useSusuki)
    p.T     = 1;
    p.E0_fd = 2.0;
    p.P     = 0.9;
  else
    % Parameters from figure 4.
    p.T     = 1.5;
    p.E0_fd = 1.6;
    % Choice of P determines which subfigure of figure 8 is in effect.
    p.P     = 1.0;
  end
  
  % Calculated parameters
  p.x     = 0.2 ./ p.valid_routes;
  p.xp    = p.x + p.xp_d;

  % This is the only parameter that is different than the ACC
  % submission.  Using the ACC value of Q0 resulted in an empty safe set,
  % so at Yoshihiko's suggestion I reduced Q0 such that there is a
  % (small) safe set.
  p.Q0    = 0.25 * p.P;
  if((p.H == 0) && (p.B == 0))
    p.Q = p.Q0;
  else
    error('DAE model is not currently set up to handle nonconstant Q');

    % The problem is that Q will now depend on state variable E, so the code in
    % the dynamics routine would have to be modified.  Not a huge deal for
    % the DAE model, but the fully ODE model would have to be rederived.
    %
    % Q = p.Q0 + p.H * E + p.B * E.^2
  end

  % Reachability parameters.  Also appears as E_c in the papers.
  p.E_min = 0.70;
    
  %---------------------------------------------------------------------------
  % The mode sequence for this scenario is that the system is operating
  % normally (two transmission routes).  A failure occurs in one route,
  % causing the system to drop down to one route for a short period (~12
  % cycles = 0.2 seconds) before an alternative second route is reopened.

  % What is a sufficiently long time that the reachable tube will have reached
  % an equilibrium?
  t_equilibrium = 6;
  
  % Defines the default state, which should be run to a fixpoint of the
  % reachable tube.
  nominal.routes = 2;
  nominal.time = t_equilibrium;

  % Defines the failure state, which only exists until the failure is
  % noticed and corrected.
  failure.routes = 1;
  failure.time = 0.2;
  
  % Defines the corrected state, in which a backup route has been
  % made available.
  backup = nominal;

  %-----------------------------------------------------------------------
  if isempty(data_filename)
    % We have no data file, so recreate the data from scratch.

    % Start from the final state, which is when the backup has been made
    % available.  Since this mode is identical to the nominal mode, we can
    % also use these results to show the behaviour with no failure (pure
    % nominal mode).  There is no outgoing switch from this mode so the
    % reach tube may only grow.
    p.num_routes = backup.routes;
    [ backup.final, g, backup.initial ] = ...
        projReachOneGen(backup.time, [], [], p, 1, accuracy);

    % We'll need an implicit surface representation of the target set later.
    target_set = backup.initial;
    
    % The middle state is in failure.  First, map the implicit surface function
    % across the reset relation.
    failure.initial = ...
        resetOneGen(backup.final, failure.routes, g, p, accuracy);

    % Now the continuous evolution.  The outgoing switch is timed, so we want
    % the reachable set (which may grow or shrink; not the reachable tube)
    % and we do not restrict the updates (although it turns out not to make
    % a difference).  For such calculations, we would normally have to also
    % take the union with the target set after each timestep; however, we
    % know that the dynamics in this case are damped to zero within the
    % target set ==> no trajectories may leave the target set ==> it is
    % sufficient to take the union with the target set just once at the end.
    p.num_routes = failure.routes;
    failure.final = ...
        projReachOneGen(failure.time, g, failure.initial, p, 0, accuracy);
    failure.final = shapeUnion(failure.final, target_set);
    
    % The first state is nominal operation before failure.  First, map
    % the implicit surface function across the reset relation.
    nominal.initial = ...
        resetOneGen(failure.final, nominal.routes, g, p, accuracy);
    
    % Now the continuous evolution.  The outgoing switch is uncontrolled, so we
    % want the reachable tube and we must restrict the updates.  You may
    % notice that the level set calculations below take fewer timesteps even
    % when working in the same mode as the backup mode.  This effect is
    % definitely linked to the initial conditions (the initial unsafe set is
    % much larger in the simulations below than in the nominal mode).  I
    % believe that it is specifically caused by the constraint that the
    % reach set may only grow; however, I haven't verified that belief.
    p.num_routes = nominal.routes;
    nominal.final = ...
        projReachOneGen(nominal.time, g, nominal.initial, p, 1, accuracy);

    % We'd also like to know what would happen if the failure were
    % detected and corrected instantly; in other words, there was no time
    % for continuous evolution in the failure mode.
    instant = nominal;
    instant.initial = ...
        resetOneGen(failure.initial, instant.routes, g, p, accuracy);
    p.num_routes = instant.routes;
    instant.final = ...
        projReachOneGen(instant.time, g, instant.initial, p, 1, accuracy);
    
  %-----------------------------------------------------------------------
  else
    % If data_filename is not empty, load the data from the file.
    
    % Load the data set.
    load(data_filename);
    
    % Rebuild the grid structure.
    g = processGrid(g);
    
  end

  %-----------------------------------------------------------------------
  % Visualization of the unsafe set.  The labeling of the various sets was
  % done by hand for the HSCC submission.
  figure;

  % We need to undo the scaling in the E_f dimension.
  pg = g.xs;
  pg{2} = 10 * pg{2};
  
  % Target set.
  [ garbage, h_target ] = contour(pg{:}, target_set, [ 0 0 ], 'b-');
  hold on;

  % Set of states that were unsafe even without the route failure.  Normally
  % we would have to run the nominal mode by itself, but we can use the
  % backup mode results here because the two modes have identical
  % parameters.
  [ garbage, h_always ] = contour(pg{:}, backup.final, [ 0 0 ], 'b-');

  % Set of states that became unsafe due to discrete transitions.
  [ garbage, h_discrete ] =  contour(pg{:}, instant.initial, [ 0 0 ], 'r--');

  % Set of states that became unsafe with instant failure detection.
  [ garbage, h_instant ] =  contour(pg{:}, instant.final, [ 0 0 ], 'k--o');

  % Set of states that became unsafe with instant failure detection.
  [ garbage, h_nominal ] =  contour(pg{:}, nominal.final, [ 0 0 ], 'm-');

  % Pretty up the plot.
  set([ h_target; h_always; h_discrete; h_instant; h_nominal ], ...
      'LineWidth', 2);
  set(gca, 'FontSize', 20);
  xlabel('E');
  ylabel('E_f');
  daspect([ 1 10 1 ]);
  axis( [ 0.6, 1.6, 0, 7 ]);
  grid on;
  
  
