function [ old_coords, new_coords ] = ...
                            switchOneGen(new_routes, show_plot, samples_E, p)
% switchOneGen: Computes and displays effects when transmission routes change
%
%   [ old_coords, new_coords ] = ...
%                           switchOneGen(new_routes, show_plot, samples_E, p)
%
% This function uses a simple 3D DAE model of the power system from:
%
%   Venkatasubramanian, Schattler & Jaborszky, "Voltage Dynamics: Study
%   of a Generator with Voltage Control, Transmission, and Matched MW
%   Load," IEEE Trans. Automatic Control, v. 37, n. 11, pp. 1717-1733
%   (November, 1992).
%
% The hybrid system used to model transmission route changes is adapted from
%
%   Susuki & Hikihara, "Predicting Voltage Instability of Power System via
%   Hybrid System Reachability Analysis," Proceedings of the 2007 American
%   Control Conference, New York, NY, pp. 4166 - 4171 (July 2007).
%  
% This particular file computes the continuous state reset that occurs when
% the number of transmission routes changes due to relay switching.  The
% continuous state must reset because the DAE constraint changes when the
% number of transmission routes changes.  At the suggestion of Susuki, we
% assume that the DAE constraint actually represents an extremely fast
% timescale ODE, and so during a discrete transition of the hybrid system we
% allow the algebraic variable in the DAE (in this case E) to move to the
% new constraint surface while holding the differential variables (E' and
% E_fd) constant.  It is not clear how physically realistic this
% treatment is.
%
% Note that because the algebraic constraint is constant with respect to
% E_fd, this routine ignores the third dimension and works only with the E
% and E' state variables.
%
% The algebraic constraint can be represented as a function 
%
%   E'(E) = (1/E) sqrt((x' P)^2 + (X' Q_0 + E^2)^2)
%
% for the case where H = 0 = B.  A slightly more complicated formula can
% be determined for the more general case.
%
% Input Parameters:
%
%   new_routes: Integer in { 1, 2 }.  The number of transmission routes in
%   the system after the switch.  The other value is chosen as the number of
%   routes before the switch.  Optional.  Default = 2.
%
%   show_plot: Boolean.  Show a 2D plot (in E vs E' space) of the switching
%   behaviour.  The two constraints are shown as contours (blue solid line
%   for the new transmission line(s), red dashed line for the old
%   transmission line(s)), while the sampled switching points are shown as
%   symbols connected by black dotted lines.  Optional.  Default = 1 (show
%   the plot).
%
%   samples_E: Column vector.  Values of E at which to sample the transition
%   after the switch.  Values for E' are determined by the function given
%   above.  Optional.  Default is to use the E values in the grid on which
%   the contour plot is generated.  Use empty vector [] as a placeholder
%   if you would like the default.
%
%   p: Structure.  Parameter structure for the electrical system model.
%   Note that if you use something other than the default parameters, the
%   visualization may not look very good (eg bounds may be wrong).
%   Optional.  Defaults to the parameter values from figures 4 & 8 of
%   Venkatasubramanian.
%
% Output Parameters:
%
%   old_coords: 2 column matrix.  Contains the E (first column) and E'
%   (second column) values at the beginning of the switch.  Rows correspond
%   to the rows in the new_coords matrix, and to the entries in the
%   samples_E input parameter.  If the fast dynamics are unstable for a
%   particular set of new coordinates, then an NaN row indicates that no
%   suitable old coordinates could be found.
%
%   new_coords: 2 column matrix.  Contains the E (first column) and E'
%   (second column) values at the end of the switch.  Rows correspond to
%   the rows in the old_coords matrix.  Values of E are taken from the
%   samples_E input parameter.

% Copyright 2007 Ian M. Mitchell (mitchell@cs.ubc.ca).
% This software is used, copied and distributed under the licensing 
%   agreement contained in the file LICENSE in the top directory of 
%   the distribution.
%
% Ian Mitchell, 9/28/07

  % What is the maximum amount of time that we are willing to try to
  % reach the constraint?
  tspan = [ 0 1000 ];

  %---------------------------------------------------------------------------
  % Input defaults.
  
  if(nargin < 1)
    new_routes = 2;
    old_routes = 1;
  else
    if ~any(new_routes == [ 1 2 ])
      error('new_routes parameter must be 1 or 2, not %d', new_routes);
    end
    old_routes = 3 - new_routes;
  end
  
  if(nargin < 2)
    show_plot = 1;
  end

  if(nargin < 3)
    % We'll use an empty vector to signal that we need to fill this
    % variable with values from the grid.
    samples_E = [];
  else
    if(size(samples_E, 2) > 1)
      error('samples_E parameter must be a column vector.');
    end
  end

  if(nargin < 4)
    % Use an empty array to signal that we need to set up parameter values.
    p = [];
  end
  
  %---------------------------------------------------------------------------
  % Create the grid to use for visualizing the constraint surface
  % contours (and also it provides a good idea of the region of interest
  % in the plotting).  Dimensions are in the order E, E'.  We don't
  % bother with E_fd, since it is irrelevant to the constraint.
  g.dim = 2;
  g.N = [ 61; 51 ];
  g.min = [ 0.5; 0.8 ];
  g.max = [ 1.7; 1.8 ];
  g.bdry = @addGhostExtrapolate;
  g = processGrid(g);

  % Rename the grid node coordinates for easy reference.
  E = g.xs{1};
  Ep = g.xs{2};

  % If necessary, use the grid to generate the E samples for the
  % switching locations.
  if isempty(samples_E)
    samples_E = g.vs{1};
  end
  
  %---------------------------------------------------------------------------
  if isempty(p)

    p.valid_routes = [ 1; 2 ];
    
    % Fill in the model parameter values if necessary.
    p.Tp_d0 = 5;
    p.E_r   = 1.0;
    p.H     = 0;
    p.x_d   = 1.2;
    p.B     = 0;
    p.xp_d  = 0.2;
    p.K     = 7.0;

    p.T     = 1.5;
    p.E0_fd = 1.6;
    % Choice of P determines which subfigure of figure 8 is in effect.
    p.P     = 1.0;

    % Calculated parameters.
    p.x     = [ 0.2; 0.1 ];
    p.xp    = p.x + p.xp_d;
  
    p.Q0    = 0.25 * p.P;
    if((p.H == 0) && (p.B == 0))
      % By default, we will have the same Q in both old and new 
      % configurations: no extra Q factor added in.
      p.Q = p.Q0;
    else
      error('Model is not currently set up to handle nonconstant Q');

      % The problem is that Q will now depend on state variable E, so the 
      % code in the dynamics routine would have to be modified.  Not a 
      % huge deal for the DAE model, but the fully ODE model would have 
      % to be rederived.
      %
      % p.Q = p.Q0 + p.H * E + p.B * E.^2
    end
    
    % Reachability parameters.  For this function, we use this parameter to
    % declare a minimum value of E below which we will not bother to try to
    % switch (either the new or the old coordinate is inside the target set
    % and already unsafe).
    p.E_min = 0.70;
  
  end
  
  % Now some additional fields for the parameter structure specific to this
  % routine.
    
  % We need to pass the number of old routes in the parameter structure
  % so that it is available inside the ODE routines.
  p.old_routes = old_routes;

  % We also need an upper bound for E in case the switch ODE is unstable.  We
  % will use the upper bound on the grid as a convenient value.
  p.E_max = max(E(:));

  % The old state before the switch is determined when the ODE integration is
  % close enough to satisfying the old constraint.  I tried 1e-6, but that
  % seems too strict for the integrator to achieve.
  p.close_to_constraint = 1e-4;
  
  %---------------------------------------------------------------------------
  % Visualize the two constraint surfaces.

  if show_plot
    % The algebraic constraint g.
    new_constraint = (Ep.^2 .* E.^2 - (p.xp(new_routes) * p.P)^2 ...
                      - (p.xp(new_routes) * p.Q + E.^2).^2);
    old_constraint = (Ep.^2 .* E.^2 - (p.xp(old_routes) * p.P)^2 ...
                      - (p.xp(old_routes) * p.Q + E.^2).^2);
  
    % Surface visualization.
    figure;
    hold on;
    [ garbage, hNC ] = contour(E, Ep, new_constraint, [ 0 0 ], 'b-');
    [ garbage, hOC ] = contour(E, Ep, old_constraint, [ 0 0 ], 'r--');
    axis equal;
    axis(g.axis);
    grid on;
    legend([ 'new ' num2str(new_routes) ' route(s)' ], ...
           [ 'old ' num2str(old_routes) ' route(s)' ], ...
           'Location', 'Best');
    set(hNC, 'LineWidth', 3);
    set(hOC, 'LineWidth', 2);
  end
  
  %---------------------------------------------------------------------------
  % Set up the ODE which we use to find the reset relation.  If the DAE's
  % constraint is g(E, Ep) = 0, then we use the ODE dE/dt = -g(E, Ep) with
  % Ep fixed.  The negative sign takes care of the fact that we start with
  % the state at the end of the switch and go backwards.
  dynamics_func = @switchODE;

  % Set the options that are the same for all simulations.
  options = odeset('Vectorized', 'on');

  % Detects that the trajectory is getting close to either the upper or
  % lower bounds of interest, and ends the simulation.
  options = odeset(options, 'Events', @detectBounds);

  % We need accuracy better than our requirement for being close to the
  % constraint surface.
  options = odeset(options, 'AbsTol', 0.01 * p.close_to_constraint);
  
  %---------------------------------------------------------------------------
  % Initial conditions.  We are given the sample values for E at the end
  % of the switch.  Now we use the constraint surface function to
  % determine the values for Ep.
  samples_Ep = sqrt((p.xp(new_routes) * p.Q + samples_E.^2).^2 ...
                    + (p.xp(new_routes) * p.P)^2) ./ samples_E;
  
  %---------------------------------------------------------------------------
  % Set up the trajectories.
  new_coords = [ samples_E, samples_Ep ];
  % Automatically handles the case where the initial or final states are out
  % of bounds.
  old_coords = NaN * ones(size(new_coords));
  for i = 1 : length(samples_E)

    % Regular initial condition.
    y0 = [ samples_E(i); samples_Ep(i) ];

    % If this is a valid new state for E.
    if((samples_E(i) > p.E_min) && (samples_E(i) < p.E_max))

      % Give it a whirl.
      [ t, y, te, ye, ie ] = ode15s(dynamics_func, tspan, y0, options, p);
      
      if isempty(ie)
        % We reached the end of the integration time, which should not happen.
        error('Constraint surface not achieved after %d time units', ...
              tspan(end));
        
      elseif(ie == 3)
        % We are near the constraint surface.
        
        % State that caused the event is "close enough" to the constraint.
        old_coords(i, :) = ye;
      
        if show_plot
          % For plotting purposes, we won't worry about getting the event
          % causing state.
          hL = plot(y(:,1), y(:,2), 'k:');
          hO = plot(y(end,1), y(end,2), 'ko');
          if(new_routes == 1)
            % We know that the manifold for 1 route is to the left of the
            % manifold for two routes, so we'll choose our symbols to
            % look like little arrow heads.
            hN = plot(y0(1), y0(2), 'k<');
          else            
            hN = plot(y0(1), y0(2), 'k>');
          end

          % Make the lines thick enough to see.
          set(hL, 'LineWidth', 2);
          set([ hN; hO ], 'MarkerSize', 8, 'LineWidth', 2);
        end
      
      else
        % We terminated on one of the bound violation events.
        % Old state is out of bounds.
        if show_plot
          hN = plot(y(1,1), y(1,2), 'ms');
        end
        
      end
    end
    
  end

  %---------------------------------------------------------------------------
  % Pretty up the results.
  if show_plot
    set(gca, 'FontSize', 20);
    xlabel('E');  ylabel('E`');
    axis tight;
  end

  
  
%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function out = constraint(E, Ep, p)
% constraint: The constraint for the model.
%
% The state variables are E and Ep, and structure p contains model
% parameters.  The function is vectorized for E and Ep.
  
  out = (Ep.^2 .* E.^2 - (p.xp(p.old_routes) * p.P)^2 ...
            - (p.xp(p.old_routes) * p.Q + E.^2).^2);

  
  
%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function ydot = switchODE(t, y, p)
% switchODE: implements the switch ODE.
%
% p is a structure containing the system parameters.
%
% The switch ODE is simply the constraint function, where we let E change
% but not Ep: dE/dt = -g(E,Ep), dEp/dt = 0.  We introduce a minus sign
% because we are working backward from the state after the switch.
  
  % Unpack the state.
  E = y(1,:);
  Ep = y(2,:);

  % The dynamics.  E_dot is just the constraint function.
  E_dot = constraint(E, Ep, p);
  Ep_dot = zeros(size(E_dot));
  
  % Pack it all back up.
  ydot = [ E_dot; Ep_dot ];
  
  
%---------------------------------------------------------------------------
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%---------------------------------------------------------------------------
function [ value, isterminal, direction ] = detectBounds(t, y, p)
% detectBounds: Detects that the trajectory is going out of bounds.
%
% For the purposes of switching, we are not interested if the state before
% or after the switch is inside the unsafe target set or too large (which
% may indicate an unstable switching ODE).  We also want to terminate if
% we are near the constraint surface.
  
  % Unpack the relevant state vector(s).
  E = y(1,:);
  Ep = y(2,:);

  % The first two functions detect the state E going out of bounds.  The
  % final function detects proximity to the constraint.
  value = [ E - p.E_min; p.E_max - E; ...
            abs(constraint(E, Ep, p)) - p.close_to_constraint ];
  isterminal = ones(size(value));
  direction = zeros(size(value));
  
